import operator
import math
import ode
import xode.transform
from pygame.rect import Rect

def sphere_rect(self):
    r = Rect(0, 0, 2*self.radius, 2*self.radius)
    r.center = self.getBody().getPosition()[:2]
    return r

class Geometry(object):
    def __init__(self):
        self.space = ode.Space()
        self.world = ode.World()
        self.world.setERP(0.8)
        self.world.setCFM(1e-8)
        self.bodies = {}
        self.geoms = {}
        self.contactgroup = ode.JointGroup()
        
    def new_body(self, object, pos, velo, accel):
        body = ode.Body(self.world)
        body.object = object
        body.setPosition(tuple(pos)+(0,))
        body.setLinearVel(tuple(velo)+(0,))
        body.setForce(tuple(accel)+(0,))
        self.bodies[object] = body
        self.geoms[object] = {}
            
    def objects(self):
        return self.bodies.keys()

    def add_geom(self, object, label, geom, offset=(0,0)):
        geom.object = object
        if offset != (0, 0):
            t = ode.GeomTransform()
            geom.setPosition(offset+(0,))
            t.setGeom(geom)
            t.shape = 'transform'
            geom = t
        geom.object = object
        geom.setBody(self.bodies[object])
        geom.label = label
        self.space.add(geom)
        self.geoms[object][label] = geom

    def add_box(self, object, label, rect, density = 1000, rotate=0, texture=None, image=None):
        lengths = rect.size+(100,)
        b = ode.GeomBox(lengths=lengths)
        bm = self.bodies[object].getMass()
        m = ode.Mass()
        m.setBox(density, *lengths)
        # Working rotation code for when we actually implement
        # non-orthogonal walls.
        if rotate:
            trans = xode.transform.Transform()
            trans.rotate(0, 0, rotate)
            b.setRotation(trans.getRotation())
        bm.add(m)
        b.shape = 'box'
        b.size = rect.size
        b.rotate = rotate
        b.texture = texture
        b.image = image
        self.bodies[object].setMass(bm)
        self.add_geom(object, label, b, offset=rect.center)
        return b

    def add_circle(self, object, label, radius, density = 1000, center=(0,0), image=None):
        c = ode.GeomSphere(radius=radius)
        cm = self.bodies[object].getMass()
        m = ode.Mass()
        m.setSphere(density, radius)
        cm.add(m)
        c.shape = 'circle'
        c.radius = radius
        c.image = image
        c.make_rect = sphere_rect
        self.bodies[object].setMass(cm)
        self.add_geom(object, label, c)#, offset=center)

    def add_polygon(self, object, label, points):
        '''Add a polygon to an object.'''
        print "needs Trimesh, that's hard too."
        self.polys.add_new(self.objects_nicknames[object], label, points)

    def get_pos(self, object):
        '''Get the position of an object.'''
        return self.bodies[object].getPosition()[0:2]

    def set_pos(self, object, pos):
        '''Set the position of an object.'''
        self.bodies[object].setPosition(tuple(pos)+(0,))

    def get_mass(self, object):
        return self.bodies[object].getMass().mass

    def move(self, object, v):
        '''Move an object by the vector v.'''
        self.pos[self.objects_nicknames[object]] += array(v).astype(self.pos.typecode())

    def set_accel(self, object, accel):
        '''Set the acceleration of an object.'''
        self.bodies[object].setForce(tuple(accel)+(0,))

    def get_velo(self, object):
        return self.bodies[object].getLinearVel()[:2]

    def set_velo(self, object, velo):
        self.bodies[object].setLinearVel(tuple(velo)+(0,))

    def get_shapes(self):
        '''Returns all the information needed by the display to draw everything.'''
        return self.geoms

    def collide_all(self, game):
        def check(s, game):
            x, y = self.get_pos(game.player)
            if not x > 0 and not x < 0 and not y > 0 and not y < 0:
                print s, self.get_pos(game.player)
                for p, r in game.ropes.iteritems():
                    r.report(game)
                # If you ever see this message, PLEASE write me
                # If you can reproduce it, the above goes double!
                raise AssertionError, "you've been bitten by the NaN bug!"
            
        n = 10
        for i in range(n):
            check('before', game)
            for b, f in game.forces.iteritems():
                self.bodies[b].setForce(tuple(f)+(0,))
            for b, f in game.impulses.iteritems():
                f = f[0]*n, f[1]*n
                self.bodies[b].setForce(tuple(f)+(0,))
            game.impulses.clear()
            
            for p, r in game.ropes.iteritems():
                r.update(self)

            for o, b in self.bodies.iteritems():
                o.statforces = []
            check('before collide', game)
            self.space.collide((game,self.world, self.contactgroup), self.near_callback)
            check('after collide', game)
            for o, b in self.bodies.iteritems():
                if len(o.statforces) > 1:
                    #print o, b, len(o.statforces), o.statforces
                    f1 = o.statforces[0]
                    f2 = o.statforces[1]
                    dot = sum(map(operator.mul, f1, f2))
                    # rounding error on identical f1, f2. We only care
                    # if dot is close to -1.
                    if dot > 1: 
                        continue
                    #print dot
                    angle = math.degrees(math.acos(dot))
                    #print angle
                    if 165 < angle < 195:
                        # o crushed!
                        game.hit(o, "crush", None)
            check('after crush', game)
            game.flush_delete()
            check('before step', game)
            self.world.step(1.0/n)
            check('after step, %f'%(1.0/n), game)
            self.contactgroup.empty()

    def near_callback(self, args, geom1, geom2):
        """Callback function for the collide() method.
        
        This function checks if the given geoms do collide and
        creates contact joints if they do.
        """

        # Check if the objects do collide
        contacts = ode.collide(geom1, geom2)
        if (geom1.object.static) and \
           (geom2.object.static):
            return
        

        # Create contact joints
        game,world,contactgroup = args
        for c in contacts:
            (pos, normal, depth, ngeom1, ngeom2) = c.getContactGeomParams()
            o1 = geom1.object
            o2 = geom2.object
            pass1=o1.passable(part=geom1, o2=o2, n=normal, t=depth, game=game)
            pass2=o2.passable(part=geom2, o2=o1, n=normal, t=depth, game=game)
            if pass1 or pass2: return
            o1t = o1.pre_collide(part=geom1, pass1=pass1, o2=o2, pass2=pass2, n=normal, t=depth, game=game)
            o2t = o2.pre_collide(part=geom2, pass1=pass2, o2=o1, pass2=pass1, n=normal, t=depth, game=game)
            #print 'collide', normal, geom1.object, geom2.object, geom1.getPosition(), geom2.getPosition(), depth, geom1.getRotation(), geom2.getRotation()
            if o1t or o2t: return

            c.setBounce(0.9)
            c.setMu(0000)
            j = ode.ContactJoint(world, contactgroup, c)

            if not o1.static: g1 = geom1.getBody()
            else: g1 = None
            if o1.static:
                o2.statforces.append(normal)
            if not o2.static: g2 = geom2.getBody()
            else: g2 = None
            if o2.static:
                # point normal "towards" o2 (usually they point to o1)
                n = -normal[0], -normal[1], -normal[2]
                o1.statforces.append(n)
            
            if not pass2 and not pass1:
                j.attach(g1, g2)

            game.play_sound('bounce.wav')
            o1.post_collide(part=geom1, pass1=pass1, o2=o2, pass2=pass2, n=normal, t=depth, game=game)
            o2.post_collide(part=geom2, pass1=pass2, o2=o1, pass2=pass1, n=normal, t=depth, game=game)

    def delete_object(self, object):
        try:
            del self.bodies[object]
            for k, v in self.geoms[object].iteritems():
                self.space.remove(v)
            del self.geoms[object]
        except KeyError:
            # already deleted
            pass

    def delete_part(self, object, label):
        try:
            g = self.geoms[object][label]
            self.space.remove(g)
            del self.geoms[object][label]
        except KeyError:
            pass

    def contained(self, obj, rect):
        for k, v in self.geoms[obj].iteritems():
            func = globals()['%s_contained'%v.shape]
            if not func(v, rect):
                return False

        return True

def box_contained(boxg, rect):
    '''See that all polygons belonging to obj are contained in rect.'''
    x, y, z = boxg.getPosition()
    dx, dy, dz = boxg.getLengths()
    points = [(x-dx, y-dy), (x+dx, y-dy), (x+dx, y+dy), (x-dx, y+dy)]
    for p in points:
        if not rect.collidepoint(p):
            return False
    return True

def circle_contained(sphereg, rect):
    x, y, z = sphereg.getPosition()
    r = sphereg.getRadius()
    if rect.left <= x - r and rect.top <= y - r and \
           x + r < rect.right and y + r < rect.bottom:
        #Yes, contained
        return True
    else:
        return False

if __name__=="__main__":
    class Thing:
        pass
    import math
    import pygame
    g = Geometry()
    
    o1 = Thing()
    g.new_body(o1, (0,0), (0, 0), (0, 0))
    g.add_box(o1, 0, 1, pygame.Rect(0, 0, 1, 1))
##    g.add_polygon(o1, 0, [(0, 0), (60, 0), (30, 30*math.sqrt(3))])
##    o2 = Thing()
##    g.add_object(o2, (0,0), (0, 0), (0, 0), 1)

##    g.add_round(o2, 0, 16)
##    def bounce(o1, k1, o2, k2, n, t):
##        print 'bounce', o1, k1, o2, k2, n, t
##        raise BounceException
##    def overlap(o1, k1, o2, k2, n, t):
##        print 'overlap', o1, k1, o2, k2, n, t
##        g.move(o2, -n*t)        
##        raise OverlapException
        
##    try:
##        g.collide_all(overlap, bounce)
##    except OverlapException:
##        pass
    
##    print 'o1 now at:', g.get_pos(o1)
##    print 'o2 now at:', g.get_pos(o2)
##    g.move(o2, array([.01, 0]))
##    try:
##        g.collide_all(overlap, bounce)
##    except OverlapException:
##        pass


##    g.move(o2, array([25, 40]))
##    g.set_velo(o2, array([0, -1.07191]))

##    try:
##        g.collide_all(overlap, bounce)
##    except BounceException:
##        pass
##    print "Everything works!"
